# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test that predictions of space usage via different subcommands of
straits-predict-usage match.
"""

# isort: STDLIB
import json
import os
import subprocess
import unittest

# isort: THIRDPARTY
from justbytes import GiB, Range, TiB

_STRATIS_PREDICT_USAGE = os.environ["STRATIS_PREDICT_USAGE"]


def _call_predict_usage_pool(
    encrypted, device_sizes, *, fs_sizes=None, overprovision=True
):
    """
    Call stratis-predict-usage and return JSON result.

    :param bool encrypted: true if pool is to be encrypted
    :param device_sizes: list of sizes of devices for pool
    :type device_sizes: list of str
    :param fs_sizes: list of filesystem sizes
    :type fs_sizes: list of Range
    :param bool overprovision: whether it is allowed to overprovision the pool
    """
    with subprocess.Popen(
        [_STRATIS_PREDICT_USAGE, "pool"]
        + [f"--device-size={size.magnitude}" for size in device_sizes]
        + (
            []
            if fs_sizes is None
            else [(f"--filesystem-size={size.magnitude}") for size in fs_sizes]
        )
        + (["--encrypted"] if encrypted else [])
        + ([] if overprovision else ["--no-overprovision"]),
        stdout=subprocess.PIPE,
    ) as command:
        outs, errs = command.communicate()
        if command.returncode != 0:
            raise RuntimeError(
                f"Invocation of {_STRATIS_PREDICT_USAGE} returned an error: "
                f"{command.returncode,}, {errs}"
            )
        prediction = json.loads(outs)

    return prediction


def _call_predict_usage_filesystem(fs_sizes, overprovision):
    """
    Call stratis-predict-usage using filesystem subcommand.

    :param fs_specs: list of filesystem sizes
    :type fs_specs: list of Range
    :param bool overprovision: whether it is allowed to overprovision the pool
    """

    with subprocess.Popen(
        [_STRATIS_PREDICT_USAGE, "filesystem"]
        + [f"--filesystem-size={size.magnitude}" for size in fs_sizes]
        + ([] if overprovision else ["--no-overprovision"]),
        stdout=subprocess.PIPE,
    ) as command:
        outs, errs = command.communicate()
        if command.returncode != 0:
            raise RuntimeError(
                f"Invocation of {_STRATIS_PREDICT_USAGE} returned an error: "
                f"{command.returncode}, {errs}"
            )
        prediction = json.loads(outs)

    return prediction


class TestSpaceUsagePrediction(unittest.TestCase):
    """
    Test relations of filesystem prediction to pool prediction.
    """

    def test_prediction(self):
        """
        Verify that the prediction of space used by the filesystem subcommand
        is the same as the prediction obtained by computing over the results
        obtained by calling the pool subcommand with different arguments and
        taking the difference.
        """
        encrypted = False
        overprovisioned = True
        device_sizes = [Range(1, TiB)]
        fs_sizes = [Range(1, GiB)]

        pool_result_pre = _call_predict_usage_pool(
            encrypted, device_sizes, fs_sizes=None, overprovision=overprovisioned
        )
        pool_result_post = _call_predict_usage_pool(
            encrypted,
            device_sizes,
            fs_sizes=fs_sizes,
            overprovision=overprovisioned,
        )

        filesystem_result = _call_predict_usage_filesystem(
            fs_sizes, overprovision=overprovisioned
        )

        self.assertEqual(
            Range(pool_result_post["used"]) - Range(pool_result_pre["used"]),
            Range(filesystem_result["used"]),
        )

    def test_parsing(self):
        """
        Test some parsing behaviors.
        """

        with self.assertRaises(subprocess.CalledProcessError) as c_m:
            subprocess.run([_STRATIS_PREDICT_USAGE], check=True)

        exception = c_m.exception
        # Clap v3 uses 2 as the exit status when displaying help
        self.assertEqual(exception.returncode, 2)



# Copyright 2020 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Support for testing udev device discovery.
"""

# isort: STDLIB
import os
import random
import signal
import string
import subprocess
import time
import unittest
from tempfile import NamedTemporaryFile

# isort: THIRDPARTY
import dbus
import psutil
import pyudev

# isort: LOCAL
from stratisd_client_dbus import (
    Blockdev,
    Manager,
    MOPool,
    ObjectManager,
    Pool,
    StratisdErrors,
    get_object,
    pools,
)
from stratisd_client_dbus._constants import TOP_OBJECT

from ._dm import _get_stratis_devices, remove_stratis_setup
from ._loopback import LoopBackDevices

_STRATISD = os.environ["STRATISD"]

CRYPTO_LUKS_FS_TYPE = "crypto_LUKS"
STRATIS_FS_TYPE = "stratis"


def random_string(length):
    """
    Generates a random string with the prefix 'stratis_'
    :param length: Length of random part of string
    :return: String
    """
    return f'stratis_{"".join(random.choice(string.ascii_uppercase) for _ in range(length))}'


def create_pool(
    name, devices, *, key_description=None, clevis_info=None, overprovision=True
):
    """
    Creates a stratis pool.
    :param name:    Name of pool
    :param devices:  Devices to use for pool
    :param key_description: optional key description
    :type key_description: str or NoneType
    :param clevis_info: clevis information, pin and config
    :type clevis_info: pair of str * str
    :return: result of the CreatePool D-Bus method call if it succeeds
    :rtype: bool * str * list of str
    :raises RuntimeError: if pool is not created
    """
    (result, exit_code, error_str) = Manager.Methods.CreatePool(
        get_object(TOP_OBJECT),
        {
            "name": name,
            "redundancy": (True, 0),
            "devices": devices,
            "key_desc": (False, "")
            if key_description is None
            else (True, key_description),
            "clevis_info": (False, ("", ""))
            if clevis_info is None
            else (True, clevis_info),
        },
    )

    if exit_code != StratisdErrors.OK:
        raise RuntimeError(
            f"Unable to create a pool {name} with devices {devices}: {error_str}"
        )

    (_, (pool_object_path, _)) = result

    if not overprovision:
        Pool.Properties.Overprovisioning.Set(get_object(pool_object_path), False)

    return result


def get_pools(name=None):
    """
    Returns a list of all pools found by GetManagedObjects, or a list
    of pools with names matching the specified name, if passed.
    :param name: filter for pool name
    :type name: str or NoneType
    :return: list of pool information found
    :rtype: list of (str * MOPool)
    """
    managed_objects = ObjectManager.Methods.GetManagedObjects(
        get_object(TOP_OBJECT), {}
    )

    return [
        (op, MOPool(info))
        for op, info in pools(props={} if name is None else {"Name": name}).search(
            managed_objects
        )
    ]


def get_devnodes(device_object_paths):
    """
    Get the device nodes belonging to these object paths.

    :param blockdev_object_paths: list of object paths representing blockdevs
    :type blockdev_object_paths: list of str
    :returns: a list of device nodes corresponding to the object paths
    :rtype: list of str
    """
    return [
        Blockdev.Properties.Devnode.Get(get_object(op)) for op in device_object_paths
    ]


def settle():
    """
    Wait some amount and then call udevadm settle.
    :return: None
    """
    time.sleep(2)
    subprocess.check_call(["udevadm", "settle"])


def wait_for_udev_count(expected_num):
    """
    Look for devices with ID_FS_TYPE=stratis. Check as many times as can be
    done in 10 seconds or until the number of devices found is equal to the
    number of devices expected. Always get the result of at least 1 enumeration.

    This method should be used only when it is very hard to figure the device
    nodes corresponding to the Stratis block devices.

    :param int expected_num: the number of expected Stratis devices
    :return: None
    :raises RuntimeError: if unexpected number of device nodes is found
    """
    found_num = None

    context = pyudev.Context()
    end_time = time.time() + 10.0

    while time.time() < end_time and not expected_num == found_num:
        found_num = len(
            frozenset(
                [
                    x.device_node
                    for x in context.list_devices(
                        subsystem="block", ID_FS_TYPE=STRATIS_FS_TYPE
                    )
                ]
            )
        )
        time.sleep(1)

    if expected_num != found_num:
        raise RuntimeError(
            f"Found unexpected number of devnodes: "
            f"expected number: {expected_num} != found number: {found_num}"
        )


def wait_for_udev(fs_type, expected_paths):
    """
    Look for devices with ID_FS_TYPE=fs_type. Check as many times as can be
    done in 10 seconds or until the devices found are equal to the devices
    expected. Always get the result of at least 1 enumeration.
    :param str fs_type: the type to look for ("stratis" or "crypto_LUKS")
    :param expected_paths: devnodes of paths that should belong to Stratis
    :type expected_paths: list of str
    :return: None
    :raises RuntimeError: if unexpected device nodes are found
    """
    expected_devnodes = frozenset((os.path.realpath(x) for x in expected_paths))
    found_devnodes = None

    context = pyudev.Context()
    end_time = time.time() + 10.0

    while time.time() < end_time and not expected_devnodes == found_devnodes:
        found_devnodes = frozenset(
            [
                x.device_node
                for x in context.list_devices(subsystem="block", ID_FS_TYPE=fs_type)
            ]
        )
        time.sleep(1)

    if expected_devnodes != found_devnodes:
        raise RuntimeError(
            f'Found unexpected devnodes: expected devnodes: {", ".join(expected_devnodes)} '
            f'!= found_devnodes: {", ".join(found_devnodes)}'
        )


def processes(name):
    """
    Find all process matching the given name.
    :param str name: name of process to check
    :return: sequence of psutil.Process
    """
    for proc in psutil.process_iter(["name"]):
        try:
            if proc.name() == name:
                yield proc
        except psutil.NoSuchProcess:
            pass


def remove_stratis_dm_devices():
    """
    Remove Stratis device mapper devices, fail with a runtime error if
    some have been missed.
    :raises RuntimeError: if some devices are remaining
    """
    remove_stratis_setup()
    if _get_stratis_devices() != []:
        raise RuntimeError("Some devices were not removed")


class _Service:
    """
    Start and stop stratisd.
    """

    # pylint: disable=consider-using-with
    def start_service(self):
        """
        Starts the stratisd service if it is not already started. Verifies
        that it has not exited at the time the method returns. Verifies that
        the D-Bus service is available.
        """

        settle()

        if next(processes("stratisd"), None) is not None:
            raise RuntimeError("A stratisd process is already running")

        service = subprocess.Popen(
            [_STRATISD],
            text=True,
        )

        dbus_interface_present = False
        limit = time.time() + 120.0
        while (
            time.time() <= limit
            and not dbus_interface_present
            and service.poll() is None
        ):
            try:
                get_object(TOP_OBJECT)
                dbus_interface_present = True
            except dbus.exceptions.DBusException:
                time.sleep(0.5)

        time.sleep(1)
        if service.poll() is not None:
            raise RuntimeError(
                f"Daemon unexpectedly exited with exit code {service.returncode}"
            )

        if not dbus_interface_present:
            raise RuntimeError("No D-Bus interface for stratisd found")

        self._service = service  # pylint: disable=attribute-defined-outside-init
        return self

    def stop_service(self):
        """
        Stops the stratisd daemon previously spawned.
        :return: None
        """
        self._service.send_signal(signal.SIGINT)
        self._service.wait()
        if next(processes("stratisd"), None) is not None:
            raise RuntimeError("Failed to stop stratisd service")


class KernelKey:
    """
    A handle for operating on keys in the kernel keyring. The specified keys
    will be available for the lifetime of the test when used with the Python
    with keyword and will be cleaned up at the end of the scope of the with
    block.
    """

    def __init__(self, key_descs):
        """
        Initialize a key with the provided key description and key data (passphrase).
        :param key_descs: list of key descriptions, may be empty
        :type key_descs: list of (str * bytes)
        """
        self._key_descs = key_descs

    def __enter__(self):
        """
        This method allows KernelKey to be used with the "with" keyword.
        :return: The key descriptions that can be used to access the
                 provided key data in __init__.
        :raises RuntimeError: if setting a key in the keyring through stratisd
                              fails
        """
        for (key_desc, key_data) in self._key_descs:
            with NamedTemporaryFile(mode="w") as temp_file:
                temp_file.write(key_data)
                temp_file.flush()

                with open(temp_file.name, "r", encoding="utf-8") as fd_for_dbus:
                    (_, return_code, message) = Manager.Methods.SetKey(
                        get_object(TOP_OBJECT),
                        {
                            "key_desc": key_desc,
                            "key_fd": fd_for_dbus.fileno(),
                        },
                    )

                if return_code != StratisdErrors.OK:
                    raise RuntimeError(
                        f"Setting a key using stratisd failed with an error: {message}"
                    )

        return [desc for (desc, _) in self._key_descs]

    def __exit__(self, exception_type, exception_value, traceback):
        try:
            for (key_desc, _) in reversed(self._key_descs):
                (_, return_code, message) = Manager.Methods.UnsetKey(
                    get_object(TOP_OBJECT), {"key_desc": key_desc}
                )

                if return_code != StratisdErrors.OK:
                    raise RuntimeError(
                        f"Unsetting the key using stratisd failed with an error: {message}"
                    )

        except RuntimeError as rexc:
            if exception_value is None:
                raise rexc
            raise rexc from exception_value


class ServiceContextManager:
    """
    A context manager for starting and stopping the daemon.
    """

    def __init__(self):
        self._service = _Service()

    def __enter__(self):
        self._service.start_service()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._service.stop_service()

        return False


class OptionalKeyServiceContextManager:
    """
    A service context manager that accepts an optional key
    """

    def __init__(self, *, key_spec=None):
        """
        Initialize a context manager with an optional list of keys
        :param key_spec: Key description and data for kernel keys to be added
        :type key_spec: list of (str, str) or NoneType
        """
        self._ctxt_manager = ServiceContextManager()
        self._keys = KernelKey([]) if key_spec is None else KernelKey(key_spec)

    def __enter__(self):
        """
        Chain ServiceContextManager and KernelKey __enter__ methods
        :return: list of key descriptions
        :rtype: list of str
        """
        self._ctxt_manager.__enter__()
        return self._keys.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._keys.__exit__(exc_type, exc_val, exc_tb)
        self._ctxt_manager.__exit__(exc_type, exc_val, exc_tb)


class UdevTest(unittest.TestCase):
    """
    Do some setup and teardown of loopbacked devices and what not.
    """

    def setUp(self):
        self._lb_mgr = LoopBackDevices()
        self.addCleanup(self._clean_up)

    def _clean_up(self):
        """
        Cleans up the test environment
        :return: None
        """
        stratisds = list(processes("stratisd"))
        for process in stratisds:
            process.terminate()
        psutil.wait_procs(stratisds)

        remove_stratis_dm_devices()
        self._lb_mgr.destroy_all()

    def wait_for_pools(self, expected_num, *, name=None):
        """
        Returns a list of all pools found by GetManagedObjects, or a list
        of pools with names matching the specified name, if passed.
        :param int expected_num: the number of pools expected
        :param name: filter for pool name
        :type name: str or NoneType
        :return: list of pool information found
        :rtype: list of (str * MOPool)
        """
        found_num = None

        end_time = time.time() + 10.0

        while time.time() < end_time and not expected_num == found_num:
            known_pools = get_pools(name=name)
            found_num = len(known_pools)
            time.sleep(1)

        self.assertEqual(found_num, expected_num)

        return known_pools



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test that predictions of space usage match the actual.
"""

# isort: STDLIB
import json
import os
import subprocess

# isort: THIRDPARTY
from justbytes import GiB, Range, TiB

# isort: LOCAL
from stratisd_client_dbus import (
    MOBlockDev,
    MOPool,
    ObjectManager,
    Pool,
    blockdevs,
    get_object,
    pools,
)
from stratisd_client_dbus._constants import TOP_OBJECT

from ._utils import (
    OptionalKeyServiceContextManager,
    ServiceContextManager,
    UdevTest,
    create_pool,
    random_string,
)

_STRATIS_PREDICT_USAGE = os.environ["STRATIS_PREDICT_USAGE"]
# Filesystem size limit should be no more than three times actual size
_FILESYSTEM_MULT_LIMIT = 3


def _call_predict_usage(encrypted, device_sizes, *, fs_specs=None, overprovision=True):
    """
    Call stratis-predict-usage and return JSON resut.

    :param bool encrypted: true if pool is to be encrypted
    :param device_sizes: list of sizes of devices for pool
    :type device_sizes: list of str
    :param fs_specs: list of filesystem specs
    :type fs_specs: list of str * Range
    :param bool overprovision: whether it is allowed to overprovision the pool
    """
    with subprocess.Popen(
        [_STRATIS_PREDICT_USAGE, "pool"]
        + [f"--device-size={size}" for size in device_sizes]
        + (
            []
            if fs_specs is None
            else [f"--filesystem-size={size.magnitude}" for _, size in fs_specs]
        )
        + (["--encrypted"] if encrypted else [])
        + ([] if overprovision else ["--no-overprovision"]),
        stdout=subprocess.PIPE,
    ) as command:
        outs, errs = command.communicate()
        if command.returncode != 0:
            raise RuntimeError(
                f"Invocation of {_STRATIS_PREDICT_USAGE} returned an error: "
                f"{command.returncode}, {errs}"
            )
        prediction = json.loads(outs)

    return prediction


def _call_blockdev_size(dev):
    """
    Get the blockdev size for a device in bytes.
    :param str dev: device path
    :rtype: str
    """
    with subprocess.Popen(
        ["blockdev", "--getsize64", dev],
        stdout=subprocess.PIPE,
    ) as command:
        outs, _ = command.communicate()

    return outs.decode().rstrip("\n")


def _possibly_add_filesystems(pool_object_path, *, fs_specs=None):
    """
    Add filesystems to the already created pool to set up testing, if
    filesystms have been specified.

    :param str pool_object_path: the D-Bus object path
    :param fs_specs: the filesystem specs
    :type fs_specs: list of str * Range or NoneType

    :returns: the change in size of pool's TotalPhysicalUsed value
    :rtype: Range
    """
    if fs_specs is not None:
        pool_proxy = get_object(pool_object_path)

        (real, pool_used_pre) = Pool.Properties.TotalPhysicalUsed.Get(pool_proxy)
        if not real:
            raise RuntimeError("Failed to get pool usage before creating filesystems.")

        (_, return_code, message,) = Pool.Methods.CreateFilesystems(
            pool_proxy,
            {"specs": map(lambda x: (x[0], (True, str(x[1].magnitude))), fs_specs)},
        )

        if return_code != 0:
            raise RuntimeError(f"Failed to create a requested filesystem: {message}")

        (real, pool_used_post) = Pool.Properties.TotalPhysicalUsed.Get(pool_proxy)
        if not real:
            raise RuntimeError("Failed to get pool usage after creating filesystems.")

        return Range(pool_used_post) - Range(pool_used_pre)

    return Range(0)


def _get_block_device_sizes(pool_object_path, managed_objects):
    """
    Get sizes of block devices.

    :param pool_object_path: The object path of the designated pool.
    :param managed_objects: managed objects dict
    """
    modevs = [
        MOBlockDev(info)
        for objpath, info in blockdevs(props={"Pool": pool_object_path}).search(
            managed_objects
        )
    ]

    block_devices = [modev.PhysicalPath() for modev in modevs]

    return [_call_blockdev_size(dev) for dev in block_devices]


class TestSpaceUsagePrediction(UdevTest):
    """
    Test relations of prediction to reality.
    """

    def _check_fs_prediction(
        self, pre_prediction, post_prediction, change, *, overprovision=True
    ):
        """
        Check the prediction for filesystems by comparing a prediction
        before filesystems are added to one after filesystems are added.

        It should be greater or equal to the actual change but no greater than
        a statical limit times the actual change.

        :param str pre_prediction: prediction before filesystems
        :param str post_prediction: prediction after filesystems
        :param Range change: the real change in TotalPhysicalUsed result
        :param bool overprovision: whether overprovisioning is allowed on pool
        """
        pre_used = Range(pre_prediction["used"])
        post_used = Range(post_prediction["used"])

        prediction_change = post_used - pre_used

        if overprovision:
            self.assertGreaterEqual(prediction_change, change)
            self.assertLessEqual(prediction_change, _FILESYSTEM_MULT_LIMIT * change)
        else:
            # If no overprovisioning is allowed, the prediction will exceed the
            # actual used for any filesystem.
            self.assertLess(change, prediction_change)

    def _check_prediction(self, prediction, mopool):
        """
        Check the prediction against the values obtained from the D-Bus.

        :param str prediction: result of calling script, JSON format
        :param MOPool mopool: object with pool properties
        """
        encrypted = mopool.Encrypted()

        (success, total_physical_used) = mopool.TotalPhysicalUsed()
        if not success:
            raise RuntimeError("Pool's TotalPhysicalUsed property was invalid.")

        (used_prediction, total_prediction) = (
            prediction["used"],
            prediction["total"],
        )

        if encrypted:
            self.assertLess(mopool.TotalPhysicalSize(), total_prediction)
            self.assertLess(total_physical_used, used_prediction)

            diff1 = Range(total_prediction) - Range(mopool.TotalPhysicalSize())
            diff2 = Range(used_prediction) - Range(total_physical_used)

            self.assertEqual(diff1, diff2)
        else:
            self.assertEqual(mopool.TotalPhysicalSize(), total_prediction)
            self.assertEqual(total_physical_used, used_prediction)

    def _test_prediction(self, pool_name, *, fs_specs=None, overprovision=True):
        """
        Helper function to verify that the prediction matches the reality to
        an acceptable degree.

        :param str pool_name: the name of the pool to test
        :param fs_specs: filesystems to create and test
        :type fs_specs: list of of str * Range or NoneType
        :param bool overprovision: True if overprovisioning is allowed
        """
        proxy = get_object(TOP_OBJECT)

        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})

        pool_object_path, pool = next(
            pools(props={"Name": pool_name})
            .require_unique_match(True)
            .search(managed_objects)
        )

        mopool = MOPool(pool)

        physical_sizes = _get_block_device_sizes(pool_object_path, managed_objects)
        pre_prediction = _call_predict_usage(
            mopool.Encrypted(), physical_sizes, overprovision=overprovision
        )

        self._check_prediction(pre_prediction, mopool)

        change = _possibly_add_filesystems(pool_object_path, fs_specs=fs_specs)

        post_prediction = _call_predict_usage(
            mopool.Encrypted(),
            physical_sizes,
            fs_specs=fs_specs,
            overprovision=overprovision,
        )

        self._check_fs_prediction(
            pre_prediction, post_prediction, change, overprovision=overprovision
        )

    def test_prediction(self):
        """
        Verify that the prediction of space used equals the reality.
        """
        device_tokens = self._lb_mgr.create_devices(4)
        devnodes = self._lb_mgr.device_files(device_tokens)

        with ServiceContextManager():
            pool_name = random_string(5)
            create_pool(pool_name, devnodes)
            self.wait_for_pools(1)
            self._test_prediction(pool_name)

    def test_prediction_encrypted(self):
        """
        Verify that the prediction of space used equals the reality if pool
        is encrypted.
        """
        device_tokens = self._lb_mgr.create_devices(4)
        devnodes = self._lb_mgr.device_files(device_tokens)

        (key_description, key) = ("key_spec", "data")
        with OptionalKeyServiceContextManager(key_spec=[(key_description, key)]):
            pool_name = random_string(5)
            create_pool(pool_name, devnodes, key_description=key_description)
            self.wait_for_pools(1)
            self._test_prediction(pool_name)

    def test_prediction_filesystems(self):
        """
        Verify that the prediction of space used is within acceptable limits
        when creating filesystems.
        """
        device_tokens = self._lb_mgr.create_devices(4)
        devnodes = self._lb_mgr.device_files(device_tokens)

        with ServiceContextManager():
            pool_name = random_string(5)
            create_pool(pool_name, devnodes)
            self.wait_for_pools(1)
            self._test_prediction(pool_name, fs_specs=[("fs1", Range(1, TiB))])

    def test_prediction_no_overprov(self):
        """
        Verify that the prediction of space used is within acceptable limits
        when no overprovisioning is allowed.
        """
        device_tokens = self._lb_mgr.create_devices(4)
        devnodes = self._lb_mgr.device_files(device_tokens)

        with ServiceContextManager():
            pool_name = random_string(5)
            create_pool(pool_name, devnodes, overprovision=False)
            self.wait_for_pools(1)
            self._test_prediction(
                pool_name, fs_specs=[("fs1", Range(2, GiB))], overprovision=False
            )



# Copyright 2019 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test unique stratis instance.
"""


# isort: STDLIB
import os
import subprocess
import unittest

# isort: THIRDPARTY
import psutil

_STRATISD = os.environ["STRATISD"]


class TestUniqueInstance(unittest.TestCase):
    """
    Test that only one instance of stratisd can be running at any given time.
    """

    # pylint: disable=consider-using-with
    def setUp(self):
        """
        Start the original stratisd instance. Register a cleanup function to
        terminate it once started.
        """
        process = subprocess.Popen(
            [_STRATISD, "--sim"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            close_fds=True,
            env=os.environ,
        )

        while not psutil.pid_exists(process.pid):
            pass

        def cleanup():
            process.terminate()
            process.wait()

        self.addCleanup(cleanup)

    def test_unique_instance(self):
        """
        Verify that a second stratisd instance can not be started.
        """
        env = dict(os.environ)
        env["RUST_LOG"] = env.get("RUST_LOG", "") + ",nix::fcntl=debug"
        with subprocess.Popen(
            [_STRATISD, "--sim"],
            stderr=subprocess.STDOUT,
            text=True,
            close_fds=True,
            env=env,
        ) as process:
            (_, _) = process.communicate()
            self.assertEqual(process.returncode, 1)



SPECS = {
    "org.freedesktop.DBus.ObjectManager": """
<interface name="org.freedesktop.DBus.ObjectManager">
    <method name="GetManagedObjects">
      <arg name="objpath_interfaces_and_properties" type="a{oa{sa{sv}}}" direction="out" />
    </method>
  </interface>
""",
    "org.storage.stratis3.Manager.r4": """
<interface name="org.storage.stratis3.Manager.r4">
    <method name="CreatePool">
      <arg name="name" type="s" direction="in" />
      <arg name="redundancy" type="(bq)" direction="in" />
      <arg name="devices" type="as" direction="in" />
      <arg name="key_desc" type="(bs)" direction="in" />
      <arg name="clevis_info" type="(b(ss))" direction="in" />
      <arg name="result" type="(b(oao))" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="DestroyPool">
      <arg name="pool" type="o" direction="in" />
      <arg name="result" type="(bs)" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="EngineStateReport">
      <arg name="result" type="s" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="ListKeys">
      <arg name="result" type="as" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="RefreshState">
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="SetKey">
      <arg name="key_desc" type="s" direction="in" />
      <arg name="key_fd" type="h" direction="in" />
      <arg name="result" type="(bb)" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="StartPool">
      <arg name="id" type="s" direction="in" />
      <arg name="id_type" type="s" direction="in" />
      <arg name="unlock_method" type="(bs)" direction="in" />
      <arg name="result" type="(b(oaoao))" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="StopPool">
      <arg name="pool" type="o" direction="in" />
      <arg name="result" type="(bs)" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="UnsetKey">
      <arg name="key_desc" type="s" direction="in" />
      <arg name="result" type="b" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <property name="StoppedPools" type="a{sa{sv}}" access="read" />
    <property name="Version" type="s" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const" />
    </property>
  </interface>
""",
    "org.storage.stratis3.Report.r4": """
<interface name="org.storage.stratis3.Report.r4">
    <method name="GetReport">
      <arg name="name" type="s" direction="in" />
      <arg name="result" type="s" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
  </interface>
""",
    "org.storage.stratis3.blockdev.r4": """
<interface name="org.storage.stratis3.blockdev.r4">
    <property name="Devnode" type="s" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const" />
    </property>
    <property name="HardwareInfo" type="(bs)" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const" />
    </property>
    <property name="InitializationTime" type="t" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const" />
    </property>
    <property name="NewPhysicalSize" type="(bs)" access="read" />
    <property name="PhysicalPath" type="s" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const" />
    </property>
    <property name="Pool" type="o" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const" />
    </property>
    <property name="Tier" type="q" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="false" />
    </property>
    <property name="TotalPhysicalSize" type="s" access="read" />
    <property name="UserInfo" type="(bs)" access="readwrite" />
    <property name="Uuid" type="s" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const" />
    </property>
  </interface>
""",
    "org.storage.stratis3.filesystem.r4": """
<interface name="org.storage.stratis3.filesystem.r4">
    <method name="SetName">
      <arg name="name" type="s" direction="in" />
      <arg name="result" type="(bs)" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <property name="Created" type="s" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const" />
    </property>
    <property name="Devnode" type="s" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="invalidates" />
    </property>
    <property name="Name" type="s" access="read" />
    <property name="Pool" type="o" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const" />
    </property>
    <property name="Size" type="s" access="read" />
    <property name="Used" type="(bs)" access="read" />
    <property name="Uuid" type="s" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const" />
    </property>
  </interface>
""",
    "org.storage.stratis3.pool.r4": """
<interface name="org.storage.stratis3.pool.r4">
    <method name="AddCacheDevs">
      <arg name="devices" type="as" direction="in" />
      <arg name="results" type="(bao)" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="AddDataDevs">
      <arg name="devices" type="as" direction="in" />
      <arg name="results" type="(bao)" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="BindClevis">
      <arg name="pin" type="s" direction="in" />
      <arg name="json" type="s" direction="in" />
      <arg name="results" type="b" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="BindKeyring">
      <arg name="key_desc" type="s" direction="in" />
      <arg name="results" type="b" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="CreateFilesystems">
      <arg name="specs" type="a(s(bs))" direction="in" />
      <arg name="results" type="(ba(os))" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="DestroyFilesystems">
      <arg name="filesystems" type="ao" direction="in" />
      <arg name="results" type="(bas)" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="GrowPhysicalDevice">
      <arg name="dev" type="s" direction="in" />
      <arg name="results" type="b" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="InitCache">
      <arg name="devices" type="as" direction="in" />
      <arg name="results" type="(bao)" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="RebindClevis">
      <arg name="results" type="b" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="RebindKeyring">
      <arg name="key_desc" type="s" direction="in" />
      <arg name="results" type="b" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="SetName">
      <arg name="name" type="s" direction="in" />
      <arg name="result" type="(bs)" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="SnapshotFilesystem">
      <arg name="origin" type="o" direction="in" />
      <arg name="snapshot_name" type="s" direction="in" />
      <arg name="result" type="(bo)" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="UnbindClevis">
      <arg name="results" type="b" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <method name="UnbindKeyring">
      <arg name="results" type="b" direction="out" />
      <arg name="return_code" type="q" direction="out" />
      <arg name="return_string" type="s" direction="out" />
    </method>
    <property name="AllocatedSize" type="s" access="read" />
    <property name="AvailableActions" type="s" access="read" />
    <property name="ClevisInfo" type="(b(b(ss)))" access="read" />
    <property name="Encrypted" type="b" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const" />
    </property>
    <property name="FsLimit" type="t" access="readwrite" />
    <property name="HasCache" type="b" access="read" />
    <property name="KeyDescription" type="(b(bs))" access="read" />
    <property name="Name" type="s" access="read" />
    <property name="NoAllocSpace" type="b" access="read" />
    <property name="Overprovisioning" type="b" access="readwrite" />
    <property name="TotalPhysicalSize" type="s" access="read" />
    <property name="TotalPhysicalUsed" type="(bs)" access="read" />
    <property name="Uuid" type="s" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const" />
    </property>
  </interface>
""",
}



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test binding and unbinding in sequence with other actions.
"""

# isort: STDLIB
import json
import os

# isort: LOCAL
from stratisd_client_dbus import Pool, StratisdErrors, get_object

from ._utils import (
    OptionalKeyServiceContextManager,
    ServiceContextManager,
    UdevTest,
    create_pool,
    random_string,
)


class TestBindingAndAddingTrustedUrl(UdevTest):
    """
    Test binding to a tang server with a trusted URL and subsequently
    adding data devices in various orders.
    """

    _TANG_URL = os.getenv("TANG_URL")
    _CLEVIS_CONFIG = {"url": _TANG_URL, "stratis:tang:trust_url": True}
    _CLEVIS_CONFIG_STR = json.dumps(_CLEVIS_CONFIG)

    def test_binding_and_adding(self):
        """
        Verify that binding and adding succeeds when url is trusted.
        """
        device_tokens = self._lb_mgr.create_devices(2)
        initial_devnodes = self._lb_mgr.device_files(device_tokens)

        device_tokens = self._lb_mgr.create_devices(2)
        added_devnodes = self._lb_mgr.device_files(device_tokens)

        (key_description, key) = ("key_spec", "data")
        with OptionalKeyServiceContextManager(key_spec=[(key_description, key)]):
            pool_name = random_string(5)
            (_, (pool_object_path, _)) = create_pool(
                pool_name, initial_devnodes, key_description=key_description
            )
            self.wait_for_pools(1)

            (_, exit_code, message) = Pool.Methods.BindClevis(
                get_object(pool_object_path),
                {"pin": "tang", "json": self._CLEVIS_CONFIG_STR},
            )

            self.assertEqual(exit_code, StratisdErrors.OK, message)

            (_, exit_code, message) = Pool.Methods.AddDataDevs(
                get_object(pool_object_path), {"devices": added_devnodes}
            )

            self.assertEqual(exit_code, StratisdErrors.OK, message)

    def test_binding_unbinding_adding(self):
        """
        Test that binding, unbinding, and then adding devices works.
        """
        device_tokens = self._lb_mgr.create_devices(2)
        initial_devnodes = self._lb_mgr.device_files(device_tokens)

        device_tokens = self._lb_mgr.create_devices(2)
        added_devnodes = self._lb_mgr.device_files(device_tokens)

        (key_description, key) = ("key_spec", "data")
        with OptionalKeyServiceContextManager(key_spec=[(key_description, key)]):
            pool_name = random_string(5)
            (_, (pool_object_path, _)) = create_pool(
                pool_name, initial_devnodes, key_description=key_description
            )
            self.wait_for_pools(1)

            (_, exit_code, message) = Pool.Methods.BindClevis(
                get_object(pool_object_path),
                {"pin": "tang", "json": self._CLEVIS_CONFIG_STR},
            )

            self.assertEqual(exit_code, StratisdErrors.OK, message)

            (_, exit_code, message) = Pool.Methods.UnbindClevis(
                get_object(pool_object_path), {}
            )

            self.assertEqual(exit_code, StratisdErrors.OK, message)

            (_, exit_code, message) = Pool.Methods.AddDataDevs(
                get_object(pool_object_path), {"devices": added_devnodes}
            )

            self.assertEqual(exit_code, StratisdErrors.OK, message)

    def test_swap_binding(self):
        """
        Test that binding with clevis, unbinding with keyring, and then
        adding devices works.
        """
        device_tokens = self._lb_mgr.create_devices(2)
        initial_devnodes = self._lb_mgr.device_files(device_tokens)

        device_tokens = self._lb_mgr.create_devices(2)
        added_devnodes = self._lb_mgr.device_files(device_tokens)

        (key_description, key) = ("key_spec", "data")
        with OptionalKeyServiceContextManager(key_spec=[(key_description, key)]):
            pool_name = random_string(5)
            (_, (pool_object_path, _)) = create_pool(
                pool_name, initial_devnodes, key_description=key_description
            )
            self.wait_for_pools(1)

            (_, exit_code, message) = Pool.Methods.BindClevis(
                get_object(pool_object_path),
                {"pin": "tang", "json": self._CLEVIS_CONFIG_STR},
            )

            self.assertEqual(exit_code, StratisdErrors.OK, message)

            (_, exit_code, message) = Pool.Methods.UnbindKeyring(
                get_object(pool_object_path), {}
            )

            self.assertEqual(exit_code, StratisdErrors.OK, message)

            (_, exit_code, message) = Pool.Methods.AddDataDevs(
                get_object(pool_object_path), {"devices": added_devnodes}
            )

            self.assertEqual(exit_code, StratisdErrors.OK, message)

    def test_swap_binding_2(self):
        """
        Test that binding with keyring, unbinding with clevis, and then
        adding devices works.
        """
        device_tokens = self._lb_mgr.create_devices(2)
        initial_devnodes = self._lb_mgr.device_files(device_tokens)

        device_tokens = self._lb_mgr.create_devices(2)
        added_devnodes = self._lb_mgr.device_files(device_tokens)

        (key_description, key) = ("key_spec", "data")
        with OptionalKeyServiceContextManager(key_spec=[(key_description, key)]):
            clevis_info = ("tang", self._CLEVIS_CONFIG_STR)

            pool_name = random_string(5)
            (_, (pool_object_path, _)) = create_pool(
                pool_name, initial_devnodes, clevis_info=clevis_info
            )
            self.wait_for_pools(1)

            (_, exit_code, message) = Pool.Methods.BindKeyring(
                get_object(pool_object_path), {"key_desc": key_description}
            )

            self.assertEqual(exit_code, StratisdErrors.OK, message)

            (_, exit_code, message) = Pool.Methods.UnbindClevis(
                get_object(pool_object_path), {}
            )

            self.assertEqual(exit_code, StratisdErrors.OK, message)

            (_, exit_code, message) = Pool.Methods.AddDataDevs(
                get_object(pool_object_path), {"devices": added_devnodes}
            )

            self.assertEqual(exit_code, StratisdErrors.OK, message)

    def test_rebind_with_clevis(self):
        """
        Test that binding with clevis on creation, then rebinding, and then
        adding data devices works.
        """
        device_tokens = self._lb_mgr.create_devices(2)
        initial_devnodes = self._lb_mgr.device_files(device_tokens)

        device_tokens = self._lb_mgr.create_devices(2)
        added_devnodes = self._lb_mgr.device_files(device_tokens)

        with ServiceContextManager():
            pool_name = random_string(5)

            clevis_info = ("tang", self._CLEVIS_CONFIG_STR)

            (_, (pool_object_path, _)) = create_pool(
                pool_name, initial_devnodes, clevis_info=clevis_info
            )
            self.wait_for_pools(1)

            (_, exit_code, message) = Pool.Methods.RebindClevis(
                get_object(pool_object_path), {}
            )

            self.assertEqual(exit_code, StratisdErrors.OK, message)

            (_, exit_code, message) = Pool.Methods.AddDataDevs(
                get_object(pool_object_path), {"devices": added_devnodes}
            )

            self.assertEqual(exit_code, StratisdErrors.OK, message)


class TestBindingAndRebindingKernelKeyring(UdevTest):
    """
    Test binding and rebinding with a key in the kernel keyring and adding
    data devices.
    """

    def test_rebind_with_new_key_description(self):
        """
        Test that binding with key on creation, rebinding with different key,
        and then adding data devices works.
        """
        device_tokens = self._lb_mgr.create_devices(2)
        initial_devnodes = self._lb_mgr.device_files(device_tokens)

        device_tokens = self._lb_mgr.create_devices(2)
        added_devnodes = self._lb_mgr.device_files(device_tokens)

        keys = [("key_desc_1", "key_data_1"), ("key_desc_2", "key_data_2")]
        with OptionalKeyServiceContextManager(key_spec=keys):
            pool_name = random_string(5)
            (_, (pool_object_path, _)) = create_pool(
                pool_name, initial_devnodes, key_description=keys[0][0]
            )
            self.wait_for_pools(1)

            (_, exit_code, message) = Pool.Methods.RebindKeyring(
                get_object(pool_object_path), {"key_desc": keys[1][0]}
            )

            self.assertEqual(exit_code, StratisdErrors.OK, message)

            (_, exit_code, message) = Pool.Methods.AddDataDevs(
                get_object(pool_object_path), {"devices": added_devnodes}
            )

            self.assertEqual(exit_code, StratisdErrors.OK, message)
